# -*- coding: utf-8 -*-
#
# File: Searcher.py
#
# Copyright (c) 2009 by PloneGov
#
# GNU General Public License (GPL)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
# ------------------------------------------------------------------------------
from appy.gen.utils import Keywords
from Products.CMFPlone.PloneBatch import Batch
from Products.PloneMeeting.utils import getDateFromRequest

# ------------------------------------------------------------------------------
class Searcher:
    '''The searcher creates and executes queries in the portal_catalog
       which are triggered by the user from the "advanced search" screen in
       PloneMeeting.'''
    def __init__(self, meetingConfig, searchedType, sortKey, sortOrder,
                 filterKey, filterValue):
        self.meetingConfig = meetingConfig
        self.portalCatalog = meetingConfig.portal_catalog
        self.tool = meetingConfig.getParentNode()
        self.rq = meetingConfig.REQUEST # The Zope REQUEST object.
        self.searchedType = searchedType
        self.sortKey = sortKey
        self.sortOrder = sortOrder
        self.filterKey = filterKey
        self.filterValue = filterValue
        self.searchParams = self.rq.SESSION['searchParams']
        self.keywords = None

    def getMultiValue(self, paramName):
        '''Gets a multi-valued element.'''
        res = self.searchParams.get(paramName, [])
        if isinstance(res, basestring): res = [res]
        return res

    def addKeywords(self, res, fieldName):
        '''Adds self.keywords to the search parameters p_res for the field
           named p_fieldName. p_res[p_fieldName] may already contain a Keywords
           instance if a filter has been defined. In that case, we must merge
           the keywords.'''
        if res.has_key(fieldName) and isinstance(res[fieldName], Keywords):
            # Merge the keywords
            res[fieldName].merge(self.keywords)
        else:
            res[fieldName] = self.keywords

    def getItemSearchParams(self, mainParams, dateInterval):
        '''Adds to dict p_mainParams the parameters which are specific for
           performing (an) item-specific query(ies) in the portal_catalog.'''
        res = mainParams.copy()
        res['portal_type'] = self.meetingConfig.getItemTypeName()
        res['created'] = {'query': dateInterval, 'range':'minmax'}
        res['sort_on'] = self.sortKey or 'created'
        if self.keywords:
            # What fields need to be queried?
            kTargets = self.getMultiValue('item_keywords_target')
            if ('search_item_titles' in kTargets) and \
               ('search_item_decisions' in kTargets):
                # In this special case we will search in SearchableText only,
                # because its content corresponds to union of indexes Text,
                # Description and getDecision (and a little bit more...)
                # This is more efficient than performing several queries and
                # merging results (as done in the following cases).
                self.addKeywords(res, 'SearchableText')
            else:
                if 'search_item_titles' in kTargets:
                    # Search among item's title and description
                    self.addKeywords(res, 'Title')
                    self.addKeywords(res, 'Description')
                if 'search_item_decisions' in kTargets:
                    self.addKeywords(res, 'getDecision')
        proposingGroups = self.getMultiValue('proposingGroups')
        if proposingGroups:
            res['getProposingGroup'] = proposingGroups
        associatedGroups = tuple(self.getMultiValue('associatedGroups'))
        if associatedGroups:
            operator = self.searchParams.get('ag_operator', 'and')
            if (operator == 'or') and (len(associatedGroups) > 1):
                associatedGroups = ' OR '.join(associatedGroups)
            res['getAssociatedGroups'] = associatedGroups
        categories = self.getMultiValue('categories')
        if categories:
            res['getCategory'] = categories
        classifiers = self.getMultiValue('classifiers')
        if classifiers:
            res['getRawClassifier'] = classifiers
        return res

    def getMeetingSearchParams(self, mainParams, dateInterval):
        '''Adds to dict p_mainParams the parameters which are specific for
           performing (a) meeting-specific query(ies) in the portal_catalog.'''
        res = mainParams.copy()
        res['portal_type'] = self.meetingConfig.getMeetingTypeName()
        res['getDate'] = {'query': dateInterval, 'range':'minmax'}
        res['sort_on'] = self.sortKey or 'getDate'
        if self.keywords: self.addKeywords(res, 'Title')
        return res

    def getAnnexSearchParams(self, mainParams, dateInterval):
        '''Adds to dict p_mainParams the parameters which are specific for
           performing (an) annex-specific query(ies) in the portal_catalog.'''
        res = mainParams.copy()
        res['portal_type'] = 'MeetingFile'
        res['created'] = {'query': dateInterval, 'range':'minmax'}
        res['sort_on'] = self.sortKey or 'created'
        if self.keywords:
            # Search among annex title and content
            self.addKeywords(res, 'Title')
            if self.tool.getExtractTextFromFiles():
                self.addKeywords(res, 'indexExtractedText')
        return res

    def queryCatalog(self, params):
        '''Performs a single query catalog.'''
        # First, transform every Keywords instance into a string.
        for fieldName in params.keys():
            if isinstance(params[fieldName], Keywords):
                params[fieldName] = params[fieldName].get()
        # Then, perform the query
        # Is it a local query or a query to a distant site?
        if self.searchParams.get('search_site', '_local') != '_local':
            # Get the external application for sending him the request
            extApp = self.tool.get(self.searchParams.get('search_site'))
            # Perform the distant search
            return extApp.search(params)
        else:
            return self.portalCatalog(**params)[:params['sort_limit']]

    def getValueFromIndex(self, brain, indexName):
        '''Gets the value of index named p_indexName on a b_brain.'''
        if hasattr(brain, indexName):
            return getattr(brain, indexName)
        else:
            index = self.portalCatalog.Indexes[indexName]
            return index.getEntryForObject(brain.getRID())

    def mergeResults(self, results, sortKey):
        '''p_results contains several lists of brains that we need to merge.
           We need to take the p_sortKey into account.'''
        res = []
        moreBrains = True
        nextIndexes = [0] * len(results)
        nextCandidates = {} # ~{i_listNumber: brain}~
        while moreBrains:
            # Compute next candidates
            nextCandidates.clear()
            i = -1
            for nextIndex in nextIndexes:
                i += 1
                brainsList = results[i]
                if nextIndex < len(brainsList):
                    # There is at least one more candidate in this list
                    nextCandidates[i] = brainsList[nextIndex]
            if not nextCandidates:
                moreBrains = False
            else:
                # Compute the winner among all candidates
                winner = None
                winnerListNumber = None
                for listNumber, candidate in nextCandidates.iteritems():
                    if not winner:
                        winner = candidate
                        winnerListNumber = listNumber
                    else:
                        # Compare the current winner to this candidate
                        winnerKey = self.getValueFromIndex(winner, sortKey)
                        candidateKey = self.getValueFromIndex(candidate,sortKey)
                        # The comparison condition follows sort order
                        if self.sortOrder == 'reverse':
                            condition = winnerKey < candidateKey
                        else:
                            condition = winnerKey > candidateKey
                        if condition:
                            winner = candidate
                            winnerListNumber = listNumber
                        if winner.getRID() == candidate.getRID():
                            nextIndexes[listNumber] += 1
                # Add the winner to the result and prepare next iteration
                res.append(winner)
                nextIndexes[winnerListNumber] += 1
        return res

    def searchItems(self, params):
        '''Executes the portal_catalog search(es) for querying items, and
           returns corresponding brains.'''
        res = [] # We will begin by storing here a list of lists of brains.
        # Indeed, several queries may be performed.
        if params.has_key('Title'):
            # Execute the Title-related query
            tParams = params.copy()
            if tParams.has_key('Description'): del tParams['Description']
            if tParams.has_key('getDecision'): del tParams['getDecision']
            res.append(self.queryCatalog(tParams))
            del params['Title'] # The title has been "consumed".
        if params.has_key('Description'):
            # Execute the Description-related query
            dParams = params.copy()
            if dParams.has_key('getDecision'): del dParams['getDecision']
            res.append(self.queryCatalog(dParams))
            del params['Description'] # The description has been "consumed".
        if params.has_key('getDecision'):
            res.append(self.queryCatalog(params))
        # No result yet? Execute the single query from p_params.
        if not res:
            res.append(self.queryCatalog(params))
        if len(res) == 1: return res[0]
        else:
            sortKey = params['sort_on']
            return self.mergeResults(res, sortKey)[:params['sort_limit']]

    def searchMeetings(self, params):
        '''Executes the portal_catalog search(es) for querying meetings, and
           returns corresponding brains.'''
        return self.queryCatalog(params)

    def searchAnnexes(self, params):
        '''Executes the portal_catalog search(es) for querying annexes, and
           returns corresponding brains.'''
        res = [] # We will begin by storing here a list of lists of brains.
        # Indeed, several queries may be performed.
        if params.has_key('Title'):
            # Execute the Title-related query
            tParams = params.copy();
            if tParams.has_key('indexExtractedText'):
                del tParams['indexExtractedText']
            res.append(self.queryCatalog(tParams))
            del params['Title'] # The title has been "consumed".
            if params.has_key('indexExtractedText'):
                # Execute the extractedText-related query
                res.append(self.queryCatalog(params))
        # No result yet? Execute the single query from p_params.
        if not res:
            res.append(self.queryCatalog(params))
        if len(res) == 1: return res[0]
        else:
            sortKey = params['sort_on']
            return self.mergeResults(res, sortKey)[:params['sort_limit']]

    def run(self):
        '''Creates and executes queries and returns the result.'''
        rq = self.searchParams
        # Determine the start number
        batchStart = int(self.rq.get('b_start', 0))
        # Determine "from" and "to" dates that determine the time period for
        # the search.
        fromDate = getDateFromRequest(rq.get('from_day'),
            rq.get('from_month'), rq.get('from_year'), start=True)
        toDate = getDateFromRequest(rq.get('to_day'),
            rq.get('to_month'), rq.get('to_year'), start=False)
        # Prepare the keywords query if keywords have been entered by the user
        if rq.get('keywords', None):
            self.keywords = Keywords(rq.get('keywords'))
        # Prepare main search parameters.
        mainParams = {'sort_limit': self.tool.getMaxSearchResults(),
                      'sort_order': self.sortOrder}
        # If a filter has been defined on a field (ie the user typed some
        # keywords in a column header for further filtering the search), we take
        # it into account here.
        if self.filterKey:
            mainParams[self.filterKey] = Keywords(self.filterValue)
        # Perform the search.
        batchSize = self.tool.getMaxShownFound(self.searchedType)
        if self.searchedType == 'MeetingItem':
            params = self.getItemSearchParams(mainParams, [fromDate, toDate])
            itemBrains = self.searchItems(params)
            res = Batch(itemBrains, batchSize, batchStart, orphan=0)
        elif self.searchedType == 'Meeting':
            params = self.getMeetingSearchParams(mainParams, [fromDate, toDate])
            meetingBrains = self.searchMeetings(params)
            res = Batch(meetingBrains, batchSize, batchStart, orphan=0)
        elif self.searchedType == 'MeetingFile':
            params = self.getAnnexSearchParams(mainParams, [fromDate, toDate])
            annexBrains = self.searchAnnexes(params)
            res = Batch(annexBrains, batchSize, batchStart, orphan=0)
        return res
# ------------------------------------------------------------------------------
