# ------------------------------------------------------------------------------
from persistent.mapping import PersistentMapping
from appy.shared.xml_parser import XmlParser
from Products.PloneMeeting.profiles.migrations import Migrator
from Products.PloneMeeting.config import *
from Products.PloneMeeting.utils import replaceHtmlEntities, getMeetingUsers
import logging
logger = logging.getLogger('PloneMeeting')

oldPortlets = ('here/portlet_meetingitems/macros/portlet',
               'here/portlet_meetings/macros/portlet',
               'here/portlet_decisions/macros/portlet',
)
newPortlet = 'here/portlet_hubsessions/macros/portlet'

# SAX parser for converting XHTML of rich text fields into text ----------------
class XhtmlToText(XmlParser):
    def characters(self, content):
        if not self.res:
            self.res = u''
        self.res += content
    def endElement(self, elem):
        if elem == 'p':
            self.res += '\n'
        elif elem == 'br':
            self.res += '\n'

# The migration class ----------------------------------------------------------
class Migrate_To_1_9(Migrator):
    def __init__(self, context):
        Migrator.__init__(self, context)

    meetingAliases = {'(Default)': '', 'edit':'meeting_edit', 'gethtml': '',
                      'index.html': '', 'properties': '', 'sharing': '',
                      'view': ''}
    itemAliases    = {'(Default)': '', 'edit':'meetingitem_edit', 'gethtml': '',
                      'index.html': '', 'properties': '', 'sharing': '',
                      'view': ''}
    adviceTypes = {'positif': 'positive',
                   'positive': 'positive',
                   'remarks': 'positive_with_remarks',
                   'remarques': 'positive_with_remarks',
                   'negatif': 'negative',
                   'negative': 'negative'
                   }
    def _configureEditPages(self):
        '''Configures, in every meeting configuration, the Meeting and
           MeetingItem types for using the new edit forms 'meeting_edit' and
           'meetingitem_edit'.'''
        logger.info('Configuring new forms "meeting_edit" and ' \
                    '"meetingitem_edit"...')
        portal_types = self.tool.portal_types
        meetingMetaType = getattr(portal_types, 'Meeting')
        itemMetaType    = getattr(portal_types, 'MeetingItem')
        meetingMetaType.setMethodAliases(self.meetingAliases)
        itemMetaType.setMethodAliases(self.itemAliases)
        for meetingConfig in self.tool.objectValues('MeetingConfig'):
            meetingTypeName = meetingConfig.getMeetingTypeName()
            itemTypeName = meetingConfig.getItemTypeName()
            meetingType = getattr(portal_types, meetingTypeName)
            itemType = getattr(portal_types, itemTypeName)
            meetingType._actions = tuple(meetingMetaType._cloneActions())
            meetingType.setMethodAliases(self.meetingAliases)
            itemType._actions = tuple(itemMetaType._cloneActions())
            itemType.setMethodAliases(self.itemAliases)
        logger.info('Done.')

    def _updateMeetingItemActions(self):
        '''Updates every MeetingItem type (from every meeting config) because
           some actions have changed (one may now create annexes in items
           defined in a meeting configuration).'''
        logger.info('Updating MeetingItem actions for every MeetingItem ' \
                    'configuration-specific portal type...')
        for meetingConfig in self.tool.objectValues('MeetingConfig'):
            meetingConfig.at_post_edit_script()
        logger.info('Done.')

    def _addAdvicesDictFor(self, item):
        '''Add the advices dict for a given p_item.'''
        # As preamble, correct the field optionalAdvisers
        res = []
        for optionalAdviser in item.getOptionalAdvisers():
            elem = optionalAdviser
            if optionalAdviser.endswith('_advisers'):
                logger.info('Corrected field optionalAdvisers for an item ' \
                            'with names of MeetingGroups instead of names of ' \
                            'Plone groups (_advisers).')
                elem = elem[:-9]
            res.append(elem)
        item.setOptionalAdvisers(res)
        # Add the dictionary
        if not hasattr(item.aq_base, 'advices'):
            item.advices = PersistentMapping()
        item.updateAdvices()

    def _addAdvicesDict(self):
        '''Adds the new PersistentMapping on every item that stores advices.'''
        logger.info('Adds on every item the persistent dictionary for ' \
                    'storing advices.')
        brains = self.portal.portal_catalog(meta_type='MeetingItem')
        for brain in brains:
            item = brain.getObject()
            modifDate = item.modification_date
            self._addAdvicesDictFor(item)
            item.setModificationDate(modifDate)
            # We ensure that the modification date of items has not changed
            # due to these maintenance-related changes.
        logger.info('Done. Doing so for items within meeting configurations...')
        # Update also item in the config, which are not in portal_catalog
        i = 0
        for meetingConfig in self.tool.objectValues('MeetingConfig'):
            for item in meetingConfig.getItems(usage='as_recurring_item'):
                i += 1; self._addAdvicesDictFor(item)
            for item in meetingConfig.getItems(usage='as_template_item'):
                i += 1; self._addAdvicesDictFor(item)
        logger.info('Done on %d item(s).' % i)

    def _migrateAdvices(self):
        '''Convert old-way advices into new-way advices.'''
        logger.info('Converting old-way advices into new-way advices...')
        brains = self.portal.portal_catalog(meta_type='MeetingItem')
        for brain in brains:
            item = brain.getObject()
            itemTitle = item.Title().decode('utf-8')
            modifDate = item.modification_date
            oldAdvices = item.objectValues('MeetingAdvice')
            for oldAdvice in oldAdvices:
                if not oldAdvice.getAdviserName():
                    logger.warn('Old advice has no adviser group defined ' \
                                'on item %s. It will not be converted.' % \
                                itemTitle)
                else:
                    groupId = oldAdvice.getAdviserName()[:-9]
                    if groupId not in item.advices:
                        logger.warn('Old advice from advisers %s must not ' \
                                    'be given on item %s. It will not be ' \
                                    'converted.' % (groupId, itemTitle))
                    else:
                        newAdvice = item.advices[groupId]
                        # Set the advice type
                        oldType = oldAdvice.getAgreementLevel()
                        if oldType not in self.adviceTypes:
                            logger.warn('Old advice from advisers %s on item ' \
                                        '%s has an unknown agreement level ' \
                                        '%s. We convert this to "positive.".' %\
                                        (groupId, itemTitle, oldType))
                            newType = 'positive'
                        else:
                            newType = self.adviceTypes[oldType]
                        newAdvice['type'] = newType
                        # Set the comment
                        oldComment = oldAdvice.Description().strip()
                        if oldComment: # Surround it with some root element.
                            oldComment = '<root>%s</root>' % \
                                         oldComment.decode('utf-8')
                            oldComment = replaceHtmlEntities(oldComment)
                            # Convert XHTML to text.
                            text = XhtmlToText().parse(oldComment) or ''
                            newComment = text.replace(' ', '&nbsp;').replace(
                                         '\n', '<br/>').replace('\r', '')
                            newAdvice['comment'] = newComment
                        else:
                            newAdvice['comment'] = ''
                        # Set the actor
                        newAdvice['actor'] = oldAdvice.Creator()
                        newAdvice['date'] = oldAdvice.creation_date
                        i = logger.info
                        i('Advice converted for item %s, type: %s' % \
                          (itemTitle, newAdvice['type']))
                        i('  Adviser:' + newAdvice['name'])
                        i('  Comment:')
                        i(newAdvice['comment'])
                        i('  Old comment:')
                        i(oldComment)
                        i('  Actor:' + newAdvice['actor'])
                        i('  Date:' + str(newAdvice['date']))
            # Remove old advices
            if oldAdvices:
                item.manage_delObjects([a.getId() for a in oldAdvices])
                logger.info('  Old advices were deleted.')
            # We ensure that the modification date of items has not changed
            # due to these maintenance-related changes.
            item.setModificationDate(modifDate)
        logger.info('Done.')

    def _convertConfigAdviceParams(self):
        '''Converting, in the configuration, advice-related parameters.'''
        for mConfig in self.tool.objectValues('MeetingConfig'):
            if hasattr(mConfig.aq_base, 'agreementlevels'):
                logger.info('Deleting agr. levels in config %s...' % mConfig.id)
                mConfig.manage_delObjects(['agreementlevels'])
                logger.info('Done.')
            # Converting event name 'adviceAdded' to 'adviceEdited'
            itemEvents = list(mConfig.getMailItemEvents())
            if 'adviceAdded' in itemEvents:
                logger.info('Converting mail event "adviceAdded" to ' \
                            '"adviceEdited"...')
                itemEvents.remove('adviceAdded')
                itemEvents.append('adviceEdited')
                mConfig.setMailItemEvents(itemEvents)
                logger.info('Done.')

    def _removeOldPortlets(self):
        '''Removes old PloneMeeting portlets.'''
        logger.info('Removing old PloneMeeting portlets and add the new one...')
        leftPortlets = self.portal.getProperty('left_slots')
        if leftPortlets == None: leftPortlets = []
        remainingPortlets = []
        for portlet in leftPortlets:
            if portlet not in oldPortlets:
                remainingPortlets.append(portlet)
        # Add the new portlet.
        if newPortlet not in remainingPortlets:
            remainingPortlets.insert(0, newPortlet)
        self.portal.manage_changeProperties(left_slots=tuple(remainingPortlets))
        logger.info('Done. New portlets are: %s' % str(remainingPortlets))

    def _updateMeetingUserData(self, obj, fieldNames):
        '''Updates on a meeting or item in p_obj, MeetingUser-related data from
           fields p_fieldNames.'''
        for fieldName in fieldNames:
            mUsers = getMeetingUsers(obj, fieldName,
                                     theObjects=True, includeDeleted=False)
            if mUsers:
                ids = [u.ploneUserId for u in mUsers]
                exec 'obj.set%s%s(ids)' % (fieldName[0].upper(), fieldName[1:])
                logger.info('[%s]: updated %s %s...' % \
                            (obj.Title().decode('utf-8'), fieldName, str(ids)))

    meetingUserFields = ('signatories', 'attendees', 'absents', 'excused')
    def _updateMeetings(self):
        '''Updates, on every meeting, info related to MeetingUsers which now
          have different IDs.'''
        logger.info('Updating MeetingUser data on every meeting...')
        brains = self.portal.portal_catalog(meta_type='Meeting')
        for brain in brains:
            meeting = brain.getObject()
            modifDate = meeting.modification_date
            # Update MeetingUser-related fields if relevant
            self._updateMeetingUserData(meeting, self.meetingUserFields)
            # We ensure that the modification date of annexes has not changed
            # due to these maintenance-related changes.
            meeting.setModificationDate(modifDate)
        logger.info('Done.')

    itemUserFields = ('itemSignatories', 'itemAbsents')
    def _updateItems(self):
        '''Updates, on every item, info related to MeetingUsers which now
          have different IDs.'''
        logger.info('Updating MeetingUser data on every item...')
        brains = self.portal.portal_catalog(meta_type='MeetingItem')
        for brain in brains:
            item = brain.getObject()
            modifDate = item.modification_date
            # Update MeetingUser-related fields if relevant
            self._updateMeetingUserData(item, self.itemUserFields)
            # We ensure that the modification date of annexes has not changed
            # due to these maintenance-related changes.
            item.setModificationDate(modifDate)
        logger.info('Done.')

    def _updateMeetingUsers(self):
        '''Changes the ID of every MeetingUser: ID of MeetingUser must
           correspond to ID of corresponding Plone user.'''
        logger.info('Updating IDs of every MeetingUser in MeetingConfigs...')
        for mConfig in self.tool.objectValues('MeetingConfig'):
            # Update characteristics on every meetingusers folder
            usersFolder = mConfig.meetingusers
            # Allow to create things in it
            usersFolder.setConstrainTypesMode(1)
            usersFolder.setLocallyAllowedTypes(['MeetingUser'])
            usersFolder.setImmediatelyAddableTypes([])
            # View portlet_hubsessions when inside it
            if not usersFolder.hasProperty('left_slots'):
                logger.info('Adding left_slots property on folder %s' % \
                            usersFolder.absolute_url())
                usersFolder.manage_addProperty('left_slots',
                    ['here/portlet_hubsessions/macros/portlet'], 'lines')
            for meetingUser in mConfig.meetingusers.objectValues():
                if hasattr(meetingUser.aq_base, 'ploneUserId') and \
                   meetingUser.ploneUserId and \
                   (meetingUser.ploneUserId != meetingUser.id):
                    meetingUser.setId(str(meetingUser.ploneUserId))
                meetingUser.updateMeetingUser()
                meetingUser.reindexObject()
        logger.info('Done.')

    toSelectColumns = ('annexes', 'annexesDecision', 'actions')
    def _updateMeetingConfigs(self):
        '''Updates some parameters in meeting configs.'''
        logger.info('Updating selected columns in meeting configurations...')
        for mConfig in self.tool.objectValues('MeetingConfig'):
            columns = list(mConfig.getItemsListVisibleColumns())
            for colName in self.toSelectColumns:
                if colName not in columns:
                    columns.append(colName)
            if ('toDiscuss' in mConfig.getUsedItemAttributes()) and \
               ('toDiscuss' not in columns):
                columns.insert(0, 'toDiscuss')
            mConfig.setItemsListVisibleColumns(columns)
        logger.info('Done.')

    def _updateExternalApplications(self):
        '''Converts field "notify" to "usages" for every ExternalApplication.'''
        logger.info('Updating external applications...')
        for extApp in self.tool.objectValues('ExternalApplication'):
            if extApp.notify: extApp.setUsages(['notify'])
            else:             extApp.setUsages(['import'])
            logger.info('Done for "%s".' % extApp.id)
        logger.info('Done.')

    def run(self):
        self.reinstall()
        self._configureEditPages()
        self._updateMeetingItemActions()
        self._addAdvicesDict()
        self._migrateAdvices()
        self._convertConfigAdviceParams()
        self._removeOldPortlets()
        self._updateMeetings()
        self._updateItems()
        self._updateMeetingUsers()
        self._updateMeetingConfigs()
        self._updateExternalApplications()
        self.refreshDatabase(catalogs=True, workflows=True)
        logger.info('Migration finished.')

# The migration function -------------------------------------------------------
def migrate(context):
    '''This migration function does the following things:

       0) Reinstall PloneMeeting;
       1) Updates every Meeting type (from every meeting config) with the new
          meeting_edit form and every MeetingItem type with meetingitem_edit;
       2) Updates every MeetingItem type (from every meeting config) because
          some actions have changed (one may now create annexes in items defined
          in a meeting configuration);
       3) Adds the new dictionary for storing advices;
       4) Converts old advices to new ones;
       5) Converts old advice-related stuff in the config to the new way;
       6) Remove old portlets;
       7) Updates, on every meeting, info related to MeetingUsers which now
          have different IDs;
       8) Updates, on every item, info related to MeetingUsers which now
          have different IDs;
       9) Updates IDs of MeetingUsers.
       10) Converts field "notify" to "usages" for ExternalApplication.
    '''
    Migrate_To_1_9(context).run()
# ------------------------------------------------------------------------------
